//https://www.youtube.com/BitaBit

//https://www.bitabittecnologia.com.br

//https://www.instagram.com/bitabittecnologia

//Alien Invasion Game

#include <LiquidCrystal.h>

int rs = 12; //Register select
int e = 11;  //Enable
int d4 = 5;  //Data d4
int d5 = 4;  //Data d5
int d6 = 3;  //Data d6 
int d7 = 2;  //Data d7

LiquidCrystal lcd(rs, e, d4, d5, d6, d7);

//Ship controls
int up = 8;
int down = 9;
int launch = 10;

bool launch_status_pressed = true;

int score = -1;
int available_missiles = 21;

//missile_update
//Missile update speed.
//The lower the value, the faster the update
int missile_update = 50;
unsigned long int missile_current_millis = 0;

//Minimum and maximum speed of updating the alien's position
int minimum_alien_update = 50;
int maximum_alien_update = 3000;
int alien_update = minimum_alien_update;
unsigned long int alien_current_millis = 0;

bool stop_update = false;

//Used to clean up a character 
//created in a certain position
byte clear[8] =
{
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000,
  B00000
};

class PositionAndStatus
{
  private:
    bool object_exist;
    int column_pos;
    int line_pos;

  public:
    PositionAndStatus(int new_column_pos, int new_line_pos) : column_pos(new_column_pos), line_pos(new_line_pos)
    {
      object_exist = false;
    }

    inline bool objectExist(){return object_exist;}
    inline void setObjectExistence(bool exist){ object_exist = exist;}

    inline int getColumnPos(){return column_pos;}
    inline void setColumnPos(int pos){column_pos = pos;}

    inline int getLinePos(){return line_pos;}
    inline void setLinePos(int pos){line_pos = pos;}
};

class Missile : public PositionAndStatus
{
  private:
    byte missile_base[8];

    byte missile_p1[8];
    byte missile_p2[8];

    int step;

  public:
    Missile() : PositionAndStatus(3, 0)
    {
      missile_base[0] = B00000;
      missile_base[1] = B00000;
      missile_base[2] = B00000;
      missile_base[3] = B11111;
      missile_base[4] = B11111;
      missile_base[5] = B00000;
      missile_base[6] = B00000;
      missile_base[7] = B00000;

      step = 5;
  
      for(int i=0; i<8; i++)
      {
        missile_p1[i] = missile_base[i];
        missile_p2[i] = clear[i];
      }
    }

    inline byte* getMissileBase(){return missile_base;} 
    void setMissileBase(byte base[8])
    {
      for(int i=0; i<8; i++)
      {
        missile_base[i] = base[i];
      }
    }

    inline byte* getMissileP1(){return missile_p1;} 
    void setMissileP1(byte p1[8])
    {
      for(int i=0; i<8; i++)
      {
        missile_p1[i] = p1[i];
      }
    }

    inline byte* getMissileP2(){return missile_p2;} 
    void setMissileP2(byte p2[8])
    {
      for(int i=0; i<8; i++)
      {
        missile_p2[i] = p2[i];
      }
    }

    inline int getStep(){return step;}
    inline void setStep(int new_step){step = new_step;}
};

class Alien : public PositionAndStatus
{
  private:
    byte alien_base[8];

  public:
    Alien() : PositionAndStatus(0, 0)
    {
      alien_base[0] = B10001;
      alien_base[1] = B01010;
      alien_base[2] = B01110;
      alien_base[3] = B11011;
      alien_base[4] = B11011;
      alien_base[5] = B01110;
      alien_base[6] = B01010;
      alien_base[7] = B10001;
    }

    inline byte* getAlienBase(){return alien_base;} 
    void setAlienBase(byte base[8])
    {
      for(int i=0; i<8; i++)
      {
        alien_base[i] = base[i];
      }
    }
};

class Ship : public PositionAndStatus
{
  private:
    byte ship_base[8];

  public:
    Ship() : PositionAndStatus(0, 0)
    {
      ship_base[0] = B10000;
      ship_base[1] = B11000;
      ship_base[2] = B11110;
      ship_base[3] = B11001;
      ship_base[4] = B11001;
      ship_base[5] = B11110;
      ship_base[6] = B11000;
      ship_base[7] = B10000;
    }

    inline byte* getShipBase(){return ship_base;} 
    void setShipBase(byte base[8])
    {
      for(int i=0; i<8; i++)
      {
        ship_base[i] = base[i];
      }
    }
};

Missile* missile = new Missile();
Alien* alien = new Alien();
Ship* ship = new Ship();

void missileUpdate()
{
  if((millis() - missile_current_millis) >= missile_update)
  {
    missile_current_millis = millis();
    if(missile->objectExist())
    {
      byte missile_temp[8];
      //Define and write the position for the first part of the missile.
      lcd.createChar(1, missile->getMissileP1());
      lcd.setCursor(missile->getColumnPos(), missile->getLinePos());
      lcd.write(1); 

      //Define and write the position for the second part of the missile.
      lcd.createChar(2, missile->getMissileP2());
      lcd.setCursor(missile->getColumnPos() + 1, missile->getLinePos());
      lcd.write(2); 
    
      //Update the points for the first part of the missile.
      for(int i=0; i<8; ++i)
      {
        missile_temp[i] = missile->getMissileBase()[i];
        missile_temp[i] = (missile_temp[i] >> 1);
      }
      missile->setMissileP1(missile_temp);
    
      //Update the points for the second part of the missile.
      if(missile->getStep() < 5)
      {
        for(int i=0; i<8; i++)
        {
          missile_temp[i] = missile->getMissileBase()[i];
          missile_temp[i] = (missile_temp[i] << missile->getStep());
        }
       missile->setMissileP2(missile_temp);
      }
  
      int step = missile->getStep();
      step--;
      missile->setStep(step);

      //Reset the positions to the initial states.
      if( missile->getStep() < -1 )
      {
        for(int i=0; i<8; i++)
        {
          missile_temp[i] = missile->getMissileBase()[i];
          missile_temp[i] = (missile_temp[i] >> 1);
        }
        missile->setMissileP1(missile_temp);
        missile->setMissileP2(clear);
        
        //Clear where the missile part is no longer present.
        lcd.createChar(3, clear);
        lcd.setCursor(missile->getColumnPos(), missile->getLinePos());
        lcd.write(3);

        missile->setStep(4);
        int column_pos = missile->getColumnPos();
        column_pos++;
        missile->setColumnPos(column_pos);
      }
    }
  }
}

void alienUpdate()
{
  if((millis() - alien_current_millis) >= alien_update)
  {
    //Define the random update time for the position according to the defined minimum and maximum time.
    alien_update = random(minimum_alien_update, maximum_alien_update);
    
    alien_current_millis = millis();

    //If the alien exists, update the position of the line where it is.
    //if it does not exist, set the new position of the column where it will appear
    if(alien->objectExist())
    {
      int last_line_pos = alien->getLinePos();
      
      if(last_line_pos == 0)
      {
        alien->setLinePos(1);
      }
      else
      {
        alien->setLinePos(0);
      }

      lcd.setCursor(alien->getColumnPos(), last_line_pos);
      lcd.write(3);

      lcd.setCursor(alien->getColumnPos(), alien->getLinePos());
      lcd.write(4);
    }
    else
    {
      alien->setObjectExistence(true);

      int line_pos = random(0, 2);
      int column_pos = random(4, 16);
      alien->setLinePos(line_pos);
      alien->setColumnPos(column_pos);

      lcd.setCursor(alien->getColumnPos(), alien->getLinePos());
      lcd.write(4);
    }
  }
}

void checkColision()
{
  //Checks if the alien has been hit.
  if((missile->getColumnPos() == alien->getColumnPos()) && (missile->getLinePos() == alien->getLinePos()))
  {
    missile->setObjectExistence(false);
    lcd.setCursor(missile->getColumnPos(), missile->getLinePos());
    lcd.write(3);
    lcd.setCursor(missile->getColumnPos()+1, missile->getLinePos());
    lcd.write(3);
    missile->setColumnPos(3);

    alien->setObjectExistence(false);

    scoreUpdate();
  }

  //Check that the alien has not been hit.
  if(missile->getColumnPos() == 16  || ((alien->getColumnPos() + 4) == missile->getColumnPos()))
  {
    missile->setObjectExistence(false);
    lcd.setCursor(missile->getColumnPos(), missile->getLinePos());
    lcd.write(3);
    lcd.setCursor(missile->getColumnPos()+1, missile->getLinePos());
    lcd.write(3);
          
    missile->setColumnPos(3);

    availableMissilesUpdate();
  }
}

void scoreUpdate()
{
  String value;

  score++;

  //If it reaches 100 aliens, the game is won.
  if(score == 100)
  {
    showCongratulationsMessage();
    stop_update = true;
  }
  else if(score < 10)
  {
    value = "0" + String(score);
  }
  else
  {
    value = String(score);
  }

  lcd.setCursor(0, 0);
  lcd.print(value);
}

void showGameOverMessage()
{
  delay(1000);

  lcd.clear();

  lcd.setCursor(3, 0);
  lcd.print("GAME OVER");
  delay(3000);
  lcd.clear();
  delay(500);

  lcd.setCursor(3, 0);
  lcd.print("THE EARTH");
  lcd.setCursor(0, 1);
  lcd.print("HAS BEEN INVADED");
}

void showCongratulationsMessage()
{
  delay(1000);

  lcd.clear();

  lcd.setCursor(0, 0);
  lcd.print("CONGRATULATIONS!");
  delay(3000);
  lcd.clear();
  delay(500);

  lcd.setCursor(3, 0);
  lcd.print("THE EARTH");
  lcd.setCursor(0, 1);
  lcd.print("HAS BEEN SAVED!!");
}

void availableMissilesUpdate()
{
  String value;

  available_missiles--;

  //If you lose all available missiles, you lose the game.
  if(available_missiles == 0)
  {
    stop_update = true;

    showGameOverMessage();
  }
  else if(available_missiles < 10)
  {
    value = "0" + String(available_missiles);
  }
  else
  {
    value = String(available_missiles);
  }

  lcd.setCursor(0, 1);

  lcd.print(value);
}

void showIntroduction()
{
  lcd.setCursor(2, 0);
  lcd.print("ALIEN");
  delay(1000);

  lcd.setCursor(7, 1);
  lcd.print("INVASION");

  delay(3000);

  lcd.clear();

  lcd.setCursor(2, 0);
  lcd.print("DEVELOPED BY");
  lcd.setCursor(3, 1);
  lcd.print("B1T a B1T");

  delay(3000);

  lcd.clear();
}

void setup() 
{
  Serial.begin(9600);
  lcd.begin(16, 2);

  showIntroduction();

  //Set ship controls
  pinMode(up, INPUT_PULLUP);
  pinMode(down, INPUT_PULLUP);
  pinMode(launch, INPUT_PULLUP);

  //Create ship
  lcd.createChar(0, ship->getShipBase());
  lcd.setCursor(2, ship->getLinePos());
  lcd.write((byte)0);

  //Create clear character
  lcd.createChar(3, clear);

  randomSeed(analogRead(0));

  //Create alien
  lcd.createChar(4, alien->getAlienBase());

  scoreUpdate();
  availableMissilesUpdate();
}

void loop() 
{
  if(!stop_update)
  {
    //Ship moves upwards.
    if(!digitalRead(up))
    {
      ship->setLinePos(0);
      lcd.setCursor(2, ship->getLinePos());
      lcd.write((byte)0);

      lcd.setCursor(2, 1);
      lcd.write(3);
    }
    
    //Ship moves down.
    if(!digitalRead(down))
    {
      ship->setLinePos(1);
      lcd.setCursor(2, ship->getLinePos());
      lcd.write((byte)0);

      lcd.setCursor(2, 0);
      lcd.write(3);
    }
    
    //Launches the missile, one at a time.
    if(!digitalRead(launch))
    {
      if(launch_status_pressed)
      {
        if(!missile->objectExist())
        {
          missile->setObjectExistence(true);
          missile->setLinePos(ship->getLinePos());
        }
      }
      launch_status_pressed = false;
    }
    else
    {
      launch_status_pressed = true;
    }
    
    missileUpdate();
    alienUpdate();
    checkColision();
  }
}

